/*
 *  Copyright 2005 Georgetown University
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 */

package edu.georgetown.middleware.shibboleth.idp.ha;

import java.io.IOException;
import java.util.Properties;

import org.apache.log4j.Logger;
import org.jboss.cache.CacheException;
import org.jboss.cache.ConfigureException;
import org.jboss.cache.PropertyConfigurator;
import org.jboss.cache.TreeCache;

import edu.georgetown.middleware.shibboleth.idp.ha.artifact.ReplicatedArtifactMapper;
import edu.internet2.middleware.shibboleth.common.ShibResource;
import edu.internet2.middleware.shibboleth.common.ShibResource.ResourceNotAvailableException;

/**
 * A wrapper around a JBoss TreeCache object. Responsible for initalizing the TreeCache. This class looks for a property
 * file, /hashib.properties, on the classpath. It expects to find a property called cache.config which is an absolute
 * file system path to a TreeCache configuration file.
 * 
 * @author Chad La Joie
 */
public class DataReplicationService {

    /** Logger */
    private static Logger log = Logger.getLogger(ReplicatedArtifactMapper.class);

    /** Property file name */
    private static String PROPERTIES_FILE = "/hashib.properties";

    /** Cache configuration file name */
    private static String CACHE_CONFIG_FILE_PROPERTY = "cache.config";

    /** Whether we've attempted to initialize this service or not */
    private static boolean attemptedInit = false;

    /** The singleton instance of this service */
    private static DataReplicationService replSrvc = null;

    /** The JBossCache TreeCache used to replicate data */
    private TreeCache replCache;

    /**
     * Constructor
     * 
     * @throws ConfigurationException thrown when the configuration file can not be found or parsed or the 
     * data replication service can not be started
     */
    private DataReplicationService() throws ConfigurationException {
        try {
            if (log.isInfoEnabled()) {
                log.info("Intializing service to replicate Shibboleth state data cross nodes.");
            }

            ShibResource haShibPropsFile = new ShibResource(PROPERTIES_FILE);
            Properties haShibProps = new Properties();
            haShibProps.load(haShibPropsFile.getInputStream());
            String replCacheConfigFile = haShibProps.getProperty(CACHE_CONFIG_FILE_PROPERTY);

            if (log.isDebugEnabled()) {
                log.debug("Data replication service configuration file being loaded from " + replCacheConfigFile);
            }
            initializeService(replCacheConfigFile);

            if (log.isInfoEnabled()) {
                log.info("Data replication service has been initialized and started.");
            }
        }
        catch (ResourceNotAvailableException e) {
            log.fatal("Could not locate HA-Shib property file, " + PROPERTIES_FILE + " on the class path.", e);
            throw new ConfigurationException("Could not locate HA-Shib property file, " + PROPERTIES_FILE + " on the class path.", e);
        }
        catch (IOException e) {
            log.fatal("Could not read HAShib property file to load its properties.", e);
            throw new ConfigurationException("Could not read HAShib property file to load its properties.", e);
        }
    }

    /**
     * Gets the instance of the data replication service.
     * 
     * @return the instance of the data replication service
     * @throws ConfigurationException thrown if the replication service can not be created and started
     */
    public static DataReplicationService getInstance() throws ConfigurationException {
        if (replSrvc == null && !attemptedInit) {
            attemptedInit = true;
            replSrvc = new DataReplicationService();
        }

        if (replSrvc == null) {
            throw new ConfigurationException(
                    "Unable to create data replication service, data can not be replicated to other nodes.");
        }
        return replSrvc;
    }
    
    /**
     * Shuts down the replication service.
     */
    public void stopService() {
        replCache.stopService();
    }

    /**
     * Puts data in to the cache and replicates it to all other nodes.
     * 
     * @param region the cache region
     * @param key the key for the cached object, must be serializable
     * @param value the cached object, must be serializable
     * @throws CacheException thrown if the object can not be cached, this usually occurs when the cache has
     *             communication problems
     */
    public void put(String region, Object key, Object value) throws CacheException {
        replCache.put(region, key, value);
    }

    /**
     * Gets data from the cache
     * 
     * @param region the cache region to get data from
     * @param key the key of the object to retrieve
     * @return the object corresponding to that key
     * @throws CacheException thrown if the object can not be cached, this usually occurs when the cache has
     *             communication problems
     */
    public Object get(String region, Object key) throws CacheException {
        return replCache.get(region, key);
    }

    /**
     * Initializes the services.
     * 
     * @param configFile the service configuration file
     * 
     * @throws ConfigurationException thrown when there is a problem when the configuration file can not 
     * be parsed or the data replication service can not be started because of a configuration error
     */
    private void initializeService(String configFile) throws ConfigurationException{
        try {
            replCache = new TreeCache();
            PropertyConfigurator cacheConfigurator = new PropertyConfigurator();
            cacheConfigurator.configure(replCache, configFile);
            replCache.startService();
        }catch (ConfigureException e) {
            log.fatal("Data replication configuration file has errors.", e);
            throw new ConfigurationException("Data replication configuration file has error", e);
        }
        catch (Exception e) {
            log.fatal("Unable to start data replication service.", e);
            throw new ConfigurationException("Unable to start data replication service.", e);
        }
    }
}