/*
 *  Copyright 2005 Georgetown University
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 */

package edu.georgetown.middleware.shibboleth.idp.ha.artifact;

import java.io.ByteArrayInputStream;
import java.io.Serializable;

import org.opensaml.SAMLAssertion;
import org.opensaml.SAMLException;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.xml.sax.InputSource;

import edu.internet2.middleware.shibboleth.common.ServiceProvider;
import edu.internet2.middleware.shibboleth.xml.Parser;

/**
 * This class represents a serializable form of a SAML assertion to a particular
 * service provider.
 * 
 * @author Chad La Joie
 */
public class AssertionCacheObject implements Serializable {

    /** Serial version UID */
    private static final long serialVersionUID = -8338467567244734628L;

    /** The assertion XML written to a string */
    private String assertionXML;

    /** The provider ID of the service */
    private String serviceProviderId;

    /**
     * Constructor.
     * 
     * @param assertion the SAML assertion
     * @param serviceProviderId the ID of the service provider the assertion is
     *            for
     * 
     * @throws SAMLException thrown if the assertion can not be serialized
     */
    public AssertionCacheObject(SAMLAssertion assertion, String serviceProviderId) throws SAMLException {
        assertionXML = Parser.jaxpSerialize(assertion.toDOM());
        this.serviceProviderId = serviceProviderId;
    }

    /**
     * Gets the SAML assertion
     * 
     * @return the SAML assertion
     * 
     * @throws SAMLException thrown if the SAML assertion can not be recreated
     *             from the serialized assertion
     */
    public SAMLAssertion getAssertion() throws SAMLException {
        try {
            ByteArrayInputStream docIS = new ByteArrayInputStream(assertionXML.getBytes());
            Document assertionDocument = Parser.loadDom(new InputSource(docIS), false);
            
            // We have to do a clone because Xerces bombs because it won't adopt deferred nodes
            // and cloning touches each node so none will be deferred.
            // TODO remove this when Xerces fixes bug XERCESJ-1106
            Element assertionElement = (Element) assertionDocument.getDocumentElement().cloneNode(true);
            return new SAMLAssertion(assertionElement);
        } catch (Exception e) {
            return null;
        }
    }
    
    /**
     * Gets the SAML assertion as a string
     * 
     * @return the SAML assertion
     */
    public String getAssertionAsString() {
        return assertionXML;
    }

    /**
     * Gets the service provider the assertion is for
     * 
     * @return the service provider the assertion is for
     */
    public ServiceProvider getServiceProviderId() {
        return new SimpleServiceProvider(serviceProviderId);
    }

    /**
     * A simple implementation of the ServiceProvider interface
     * 
     * @author Chad La Joie
     */
    private class SimpleServiceProvider implements ServiceProvider {
        private String providerId;

        /**
         * Constructor
         * 
         * @param providerId the id of the service provider
         */
        public SimpleServiceProvider(String providerId) {
            this.providerId = providerId;
        }

        /**
         * Gets the provider ID
         */
        public String getProviderId() {
            return providerId;
        }
    }
}