/*
 *  Copyright 2005 Georgetown University
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 */

package edu.georgetown.middleware.shibboleth.idp.ha.artifact;

import org.apache.log4j.Logger;
import org.jboss.cache.CacheException;
import org.opensaml.SAMLAssertion;
import org.opensaml.SAMLException;
import org.opensaml.artifact.Artifact;
import org.w3c.dom.Element;

import edu.georgetown.middleware.shibboleth.idp.ha.ConfigurationException;
import edu.georgetown.middleware.shibboleth.idp.ha.DataReplicationService;
import edu.internet2.middleware.shibboleth.artifact.ArtifactMapping;
import edu.internet2.middleware.shibboleth.artifact.provider.BaseArtifactMapper;
import edu.internet2.middleware.shibboleth.common.ShibbolethConfigurationException;

/**
 * An IdP ArtifactMapper that replicates it's state across any number of nodes
 * to prevent loss of service is a node goes down.
 * 
 * All artifact mapper data is replicated in a the '/shibboleth/artifact' cache
 * region.
 * 
 * @author Chad La Joie
 */
public class ReplicatedArtifactMapper extends BaseArtifactMapper {

    /** Logger */
    private static Logger log = Logger.getLogger(ReplicatedArtifactMapper.class);

    /** The TreeCache node for artifact information */
    private static String CACHE_NODE = "/shibboleth/artifact";

    /** The data replication service */
    private DataReplicationService artifactCache;

    /**
     * Constructor.
     * 
     * @param config the mapper configure element from the IdP configuration
     *            file
     * @throws ShibbolethConfigurationException thrown if the data replication
     *             service can be started either because the config file is
     *             missing or because it's incorrect
     */
    public ReplicatedArtifactMapper(Element config) throws ShibbolethConfigurationException {
        super(config);

        try{
            artifactCache = DataReplicationService.getInstance();
        }catch (ConfigurationException e){
            throw new ShibbolethConfigurationException("Unable to configure or start artifact mapping replication service: " + e.getMessage());
        }
    }

    /**
     * Adds an assertion, using the artifact (in hex string form) as the key, to
     * a replicated Map.
     * 
     * @param artifact the artifact representing the assertion
     * @param mapping the assertion bundled with other useful information
     */
    public void addAssertionImpl(Artifact artifact, ArtifactMapping mapping) {
        try {
            String artifactString = artifact.toString();
            SAMLAssertion assertion = mapping.getAssertion();
            String serviceProviderId = mapping.getServiceProviderId();
            AssertionCacheObject cacheObject = new AssertionCacheObject(assertion, serviceProviderId);

            if (log.isDebugEnabled()) {
                log.debug("Replicating artifact/assertion pair associated with service " + serviceProviderId
                        + "\nArtifact: " + artifactString + "\nAssertion: " + cacheObject.getAssertionAsString());
            }
            artifactCache.put(CACHE_NODE, artifactString, cacheObject);
        } catch (SAMLException e) {
            log.error("Unable to serialize SAML assertion in preparation for artifact information replication");
        } catch (CacheException e) {
            log.error("Unable to store artifact information in cache, received the following error: " + e.getMessage());
        }
    }

    /**
     * Recovers an assertion associated with the given artifact.
     * 
     * @param artifact the artifact representing the assertion
     * 
     * @return the assertion bundled with other useful information or null if
     *         the data can not be retrieved, perhaps because it's expired and
     *         been evicted from the cache
     */
    public ArtifactMapping recoverAssertion(Artifact artifact) {
        String artifactString = artifact.toString();
        AssertionCacheObject cacheObject = null;
        ArtifactMapping artifactMapping = null;

        try {
            if (log.isDebugEnabled()) {
                log.debug("Retrieving SAML assertion for artifact " + artifactString + " from replicated cache.");
            }
            cacheObject = (AssertionCacheObject) artifactCache.get(CACHE_NODE, artifactString);

            if (cacheObject == null) {
                if (log.isInfoEnabled()) {
                    log.info("Could not retrieve SAML assertion associated with artifact " + artifactString
                            + ", the assertion has probably expired.");
                }
            } else {
                artifactMapping = new ArtifactMapping(artifact, cacheObject.getAssertion(), cacheObject
                        .getServiceProviderId());
                if (log.isDebugEnabled()) {
                    log.debug("Artifact " + artifactString + " resolved to SAML assertion associated with service "
                            + artifactMapping.getServiceProviderId() + "\nAssertion: " + cacheObject.getAssertionAsString());
                }
            }
        } catch (SAMLException e) {
            log.error("Unable to deserialize SAML assertion from replicated artifact information");
        } catch (CacheException e) {
            log.error("Unable to retrieve artifact information from cache, received the following error: "
                    + e.getMessage());
        }

        return artifactMapping;
    }
}
