/*
 *  Copyright 2005 Georgetown University
 *
 *   Licensed under the Apache License, Version 2.0 (the "License");
 *   you may not use this file except in compliance with the License.
 *   You may obtain a copy of the License at
 *
 *       http://www.apache.org/licenses/LICENSE-2.0
 *
 *   Unless required by applicable law or agreed to in writing, software
 *   distributed under the License is distributed on an "AS IS" BASIS,
 *   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *   See the License for the specific language governing permissions and
 *   limitations under the License.
 */

package edu.georgetown.middleware.shibboleth.idp.ha.nameIdentifier;

import java.lang.reflect.Constructor;
import java.security.Principal;

import org.apache.log4j.Logger;
import org.jboss.cache.CacheException;
import org.opensaml.SAMLException;
import org.opensaml.SAMLNameIdentifier;
import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import edu.georgetown.middleware.shibboleth.idp.ha.ConfigurationException;
import edu.georgetown.middleware.shibboleth.idp.ha.DataReplicationService;
import edu.internet2.middleware.shibboleth.common.IdentityProvider;
import edu.internet2.middleware.shibboleth.common.InvalidNameIdentifierException;
import edu.internet2.middleware.shibboleth.common.LocalPrincipal;
import edu.internet2.middleware.shibboleth.common.NameIdentifierMappingException;
import edu.internet2.middleware.shibboleth.common.ServiceProvider;
import edu.internet2.middleware.shibboleth.common.provider.AQHNameIdentifierMapping;

/**
 * An handle-to-principal Map whose state is replicated across server nodes to prevent loss of service if one given node
 * goes down. All handle data is replicated in a the '/shibboleth/handle' cache region.
 * 
 * @author Chad La Joie
 */
public class ReplicatedHandleMapper
        extends AQHNameIdentifierMapping {

    /** The name of the child element identifying the handle generator class */
    private static String GENERATOR_ELEMENT_NAME = "HandleGenerator";

    /** Logger */
    private static Logger log = Logger.getLogger(ReplicatedHandleMapper.class);

    /** TreeCache node id for handle information */
    private static String CACHE_NODE = "/shibboleth/handle";

    /** Data replication service */
    private DataReplicationService handleCache;

    /** The handle generator */
    private HandleGenerator handleGenerator;

    /**
     * Constructor
     * 
     * @param config the handle configure element from the IdP configuration file
     * @throws NameIdentifierMappingException thrown if no handle configuration file is specified, or if the handle
     *             replication service can not be started
     */
    public ReplicatedHandleMapper(Element config) throws NameIdentifierMappingException {
        super(config);

        try{
            handleCache = DataReplicationService.getInstance();
            initializeHandleGenerator(config);
        }catch (ConfigurationException e){
            throw new NameIdentifierMappingException("Unable to configure or start Handle mapping replication service: " + e.getMessage());
        }
    }

    /**
     * Creates a psudeonomyous handle for a given and creates a SAML name identifier with it
     * 
     * @param principal the user's principal data
     * @param sProv information about the service provider the user is interacting with
     * @param idProv information about this identity provider
     * @throws NameIdentifierMappingException thrown if there is a problem creating the handle or entering information
     *             in to the handle cache
     */
    public SAMLNameIdentifier getNameIdentifier(LocalPrincipal principal, ServiceProvider sProv, IdentityProvider idProv)
            throws NameIdentifierMappingException {

        if (principal == null) {
            log.error("A principal must be supplied to create a handle for this user.");
            throw new IllegalArgumentException("A principal must be supplied for to create a handle for this user.");
        }
        try {
            if (handleGenerator == null) {
                log.error("Attempting to use a handle generator that failed to initialize");
                throw new NameIdentifierMappingException(
                        "Attempting to use a handle generator that failed to initialize");
            }

            String handle = handleGenerator.generateHandle(principal, sProv, idProv);
            if (log.isDebugEnabled()) {
                log.debug("Assigning handle (" + handle + ") to principal (" + principal.getName() + ").");
            }

            handleCache.put(CACHE_NODE, handle, principal);

            return new SAMLNameIdentifier(handle, idProv.getProviderId(), getNameIdentifierFormat().toString());
        }
        catch (SAMLException e) {
            throw new NameIdentifierMappingException("Unable to generate handle for user: " + e);
        }
        catch (CacheException e) {
            throw new NameIdentifierMappingException(
                    "Unable to store principal/handle information in cache, received the following error: "
                            + e.getMessage());
        }
    }

    /**
     * Creates a Prinicpal from the SAML name identifier (handle) data
     * 
     * @param nameId SAML name identifier data for a user
     * @param sProv information about the service provider the user is interacting with
     * @param idProv information about this identity provider
     * @throws NameIdentifierMappingException thrown if there is a problem retrieving information from the handle cache
     * @throws InvalidNameIdentifierException thrown if there no principal associated with the handle given in the name
     *             identifier
     */
    public Principal getPrincipal(SAMLNameIdentifier nameId, ServiceProvider sProv, IdentityProvider idProv)
            throws NameIdentifierMappingException, InvalidNameIdentifierException {

        String handle = nameId.getName();
        LocalPrincipal principal;

        if (log.isDebugEnabled()) {
            log.debug("Resolving handle " + handle + " to principal ID");
        }

        verifyQualifier(nameId, idProv);

        try {
            principal = (LocalPrincipal) handleCache.get(CACHE_NODE, nameId.getName());

            if (principal == null) {
                log.error("No principal associated with handle " + handle
                        + ", the principal data has probably expired.");
                throw new InvalidNameIdentifierException(
                        "The Name Mapping Cache does not contain an entry for this Attribute Query Handle.", errorCodes);
            }
            if (log.isDebugEnabled()) {
                log.debug("Handle " + handle + " resolved to principal ID " + principal.getName());
            }
            return principal;

        }
        catch (CacheException e) {
            throw new NameIdentifierMappingException(
                    "Unable to retrieve principal information from cache, received the following error: "
                            + e.getMessage());
        }
    }

    /**
     * Reads the NameMapping configuration element, from the IdP config file, and initializes the handle generator.
     * 
     * @param config the NameMapping configuration element
     * @throws NameIdentifierMappingException thrown if there is a problem initializing the generator
     */
    private void initializeHandleGenerator(Element config) throws NameIdentifierMappingException {
        NodeList handleGeneratorElems = config.getElementsByTagName(GENERATOR_ELEMENT_NAME);
        
        //Try to initialize the user specified class
        if(handleGeneratorElems.getLength() >= 1) {
            if(handleGeneratorElems.getLength() > 1) {
                log.warn("More than one handle generator listed for replicated handle mapper, using the first one");
            }
            
            String handleGeneratorClassName = ((Element) handleGeneratorElems.item(1)).getTextContent();
            if (handleGeneratorClassName != null) {
                handleGeneratorClassName = handleGeneratorClassName.trim();
                if (handleGeneratorClassName.length() > 0) {
                    try {
                        Class generatorClazz = Class.forName(handleGeneratorClassName);
                        Class[] constructorArgs = {};
                        Constructor handleGeneratorConstructor = generatorClazz.getConstructor(constructorArgs);
                        HandleGenerator generator = (HandleGenerator) handleGeneratorConstructor
                                .newInstance(constructorArgs);
                        generator.configure(config);
                        handleGenerator = generator;
                    }
                    catch (ClassNotFoundException e) {
                        log.error("Handle generator class " + handleGeneratorClassName
                                + " could not be found on the classpath: " + e);
                        throw new NameIdentifierMappingException("Handle generator class " + handleGeneratorClassName
                                + " could not be found on the classpath");
                    }
                    catch (NoSuchMethodException e) {
                        log.error("Unable to locate public defualt constructor method for handle generator: " + e);
                        throw new NameIdentifierMappingException(
                                "Unable to locate public defualt constructor method for handle generator.");
                    }
                    catch (Exception e) {
                        log.error("Unable to instantiate handle generator: " + e);
                        throw new NameIdentifierMappingException("Unable to instantiate handle generator: ");
                    }
                } else {
                    log
                            .error("Handle generator class give in the replicated handle mapping configuration HandleGenerator element contained only whitespace");
                    throw new NameIdentifierMappingException(
                            "Handle generator class give in the replicated handle mapping configuration HandleGenerator element contained only whitespace");
                }
            }
        }
        
        //No generator was specified use the default
        if(log.isInfoEnabled()) {
            log.info("No handle generator specified, using OpenSAML default ID provider to generate handles.");
        }
        
        handleGenerator = new OpenSAMLDefaultIDHandleGenerator();
    }
}
